// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
using namespace std;


#include <fitsio.h>

#include "Dal.h"
#include "StopWatch.h"
#include "stringutil.h"
#include "dal_utilities.h"


static void
createTable(
	const string&		fileName,
	unsigned long		nRows,
	unsigned int		nCols,
	Column::DataType	type )
{
	DataSet* set = dataSetServer -> open( fileName, Dal::Create );
	Table* tab = set -> addTable( "tab", nRows );
	for( unsigned int i = 0; i < nCols; ++i )
	{
		tab -> addColumn( "col" + toString(i), type );
	} // for
	dataSetServer -> close( set );
}

struct RowRange
{
	unsigned long	startRow;
	unsigned long	stopRow;
};


static void
doSelect(
	//Table&			table,
	vector<RowRange>&	rowRangeVec,
	unsigned long		nSelects,
	unsigned long		size )
{
	unsigned long s = 0;

	for( unsigned long i = 0; i < nSelects; ++i )
	{
		RowRange	rr;
		rr.startRow = s;
		rr.stopRow = s + size;
		rowRangeVec.push_back( rr );
		s += 2 * size;
	} // for
}


static void
doFilter1(
	Table&		table,
	unsigned long	nSelects,
	unsigned long	size )
{

   vector<RowRange> rowRangeVec;

   //  do the selection
   doSelect(/*table,*/ rowRangeVec, nSelects, size );

   //  first move blocks of selected events upwards
   unsigned long toRow = 0;
   unsigned nRows = rowRangeVec.size();


   StopWatch	sw;
   for (unsigned n=0; n<nRows; ++n)
   {
      unsigned long nRows = rowRangeVec[n].stopRow -
      rowRangeVec[n].startRow + 1;
      table.copyRows(rowRangeVec[n].startRow, toRow, nRows);
      toRow += nRows;
   } // for
   cout << "copyRowsCPU:" << sw.elapsed() << endl;

   sw.reset();
   //  delete block of non-selected rows at the end
   table.deleteRows(toRow, table.rows() - toRow);
   cout << "deleteRowsCPU:" << sw.elapsed() << endl;
}


template<class T>
static void
copyRow(
	fitsfile*	fptr,
	int		columnNr,
	unsigned int	type,
	unsigned int	n,
	unsigned long	f,
	unsigned long	t,
	unsigned long	c )
{
	//T			values[ n * c ];
	T*			values = new T[ n * c ];

	fits_read_col(
			fptr,
			type,
			columnNr,
			f + 1,
			1,
			n * c,
			0,
			values,
			0,
			&status );

	fits_write_col(
			fptr,
			type,
			columnNr,
			t + 1,
			1,
			n * c,
			values,
			&status );

	fits_check( fptr );

	delete[] values;
}


static void
copyRows(
	fitsfile*	fptr,
	unsigned long	startRow,
	unsigned long	toRow,
	unsigned long	nRows )
{
	int	ncols;
	long	nrows;
	int	typecode;
	long	repeat;
	long	width;

	fits_get_num_cols( fptr, &ncols, &status );
	fits_get_num_rows( fptr, &nrows, &status );
	fits_check(fptr);

	for( int i = 0; i < ncols; ++i )
	{
		fits_get_coltype( fptr, i + 1, &typecode, &repeat, &width, &status );
		fits_check(fptr);

		switch( typecode )
		{
			case TLONG :
				copyRow<long>( fptr, i + 1, typecode, repeat, startRow, toRow, nRows );
				break;
			default :
				cerr << "error" << endl;
				exit( 0 );
		} // switch
	} // for
}


static void
deleteRows(
	fitsfile*	fptr,
	unsigned long	from,
	unsigned long	count )
{

	cout << "from:" << from << " count:" << count << endl;
	fits_delete_rows(
			fptr,
			from,
			count,
			&status );
	fits_check(fptr);
}


static void
doFilter2(
	fitsfile*	fptr,
	unsigned long	tableRows,
	unsigned long	nSelects,
	unsigned long	size )
{

   vector<RowRange> rowRangeVec;

   //  do the selection
   doSelect( rowRangeVec, nSelects, size );

   //  first move blocks of selected events upwards
   unsigned long toRow = 0;
   unsigned nRows = rowRangeVec.size();


   StopWatch	sw;
   for (unsigned n=0; n<nRows; ++n)
   {
      unsigned long nRows = rowRangeVec[n].stopRow -
      rowRangeVec[n].startRow + 1;
      copyRows(fptr, rowRangeVec[n].startRow, toRow, nRows);
      toRow += nRows;
   } // for
   cout << "copyRowsCPU:" << sw.elapsed() << endl;

   sw.reset();
   //  delete block of non-selected rows at the end
   deleteRows(fptr, toRow, tableRows - toRow);
   cout << "deleteRowsCPU:" << sw.elapsed() << endl;
}


static void
doTest1(
        const string&			master,
	const string&			fileName,
	unsigned long			tableRows,
	DataSetServer::MemoryModel	model,
	unsigned long			nSelects,
	unsigned long			size )
{
   if( !dataSetServer -> exists( "master.dat" ))
   {
      createTable( master, tableRows, 10, Column::Int32 );
   } // if

   dataSetServer -> copy( master, fileName );


   StopWatch	sw;
   DataSet* set = dataSetServer -> open( fileName, Dal::Modify, model );
   Table* tab = set -> table( 0 );
   doFilter1( *tab, nSelects, size );
   cout << "nSelects:" << nSelects << " size:" << size << " cpu:" << sw.elapsed() << endl;
   dataSetServer -> close( set );
}


static void
doTest2(
        const string&			master,
	const string&			fileName,
	unsigned long			tableRows,
	unsigned long			nSelects,
	unsigned long			size )
{
   if( !dataSetServer -> exists( "master.dat" ))
   {
      createTable( master, tableRows, 10, Column::Int32 );
   } // if

   dataSetServer -> copy( master, fileName );


   StopWatch	sw;

	fitsfile		* fptr;
	int			hdutype;
	const int		hdunum = 2;
	fits_open_file(&fptr, fileName.c_str(), READWRITE, &status );
	fits_check(fptr);

	fits_movabs_hdu(
			fptr,
			hdunum,
			&hdutype,
			&status );
	fits_check(fptr);
	long nrows;
	cout << "filename:" << fileName << ":" << fileName.c_str() << ":" << endl;
	fits_get_num_rows( fptr, &nrows, &status );
	cout << "nrows:" << nrows << endl;
   doFilter2( fptr, tableRows, nSelects, size );

   cout << "nSelects:" << nSelects << " size:" << size << " cpu:" << sw.elapsed() << endl;
	fits_close_file( fptr, &status );
	fits_check(fptr);

}


static void
test(
        const string&		master,
	const string&		fileName,
	unsigned long		tableRows,
	unsigned long		nSelects,
	unsigned long		size )
{
   cout << "HI MEMORY MODEL" << endl;
   doTest1( master, fileName, tableRows, DataSetServer::High, nSelects, size );
   //cout << "LOW MEMORY MODEL" << endl;
   //doTest1( master, fileName, tableRows, DataSetServer::HighLow, nSelects, size );
   cout << "FITS LOW MEMORY" << endl;
   doTest2( master, fileName, tableRows, nSelects, size );

}


int
main()
{
	const unsigned long tableRows = 50000;

	test( "master.dat", "test.dat", tableRows, 5, 5 );
	test( "master.dat", "test.dat", tableRows, 100, 5 );
	test( "master.dat", "test.dat", tableRows, 200, 5 );
	test( "master.dat", "test.dat", tableRows, 1000, 5 );
	test( "master.dat", "test.dat", tableRows, 4000, 5 );
}


